#pylint: disable=line-too-long
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'''
    Holds the Context object
'''
# for python3 compat
from __future__ import unicode_literals
from __future__ import absolute_import

from pyanaconda.kickstart import RepoData

from org_scientificlinux_contexts.constants import DEFAULT_CONTEXT_URL, DEFAULT_ENABLE_TESTING
from org_scientificlinux_contexts.constants import PATH_BASE, PATH_MIRROR_BASE, BASE_SUFFIX
from org_scientificlinux_contexts.constants import PATH_DEBUGINFO, PATH_MIRROR_DEBUGINFO, DEBUGINFO_SUFFIX
from org_scientificlinux_contexts.constants import PATH_SRPMS, PATH_MIRROR_SRPMS, SRPMS_SUFFIX
from org_scientificlinux_contexts.constants import PATH_TESTING, PATH_MIRROR_TESTING, TESTING_SUFFIX
from org_scientificlinux_contexts.constants import GPG_PREFIX, YUMREPO_SKIP_IF_UNAVAILABLE, REPO_PACKAGE_PREFIX

class Context(object):
    '''
        A context is a defined format yum repo with specific attributes
    '''
    def __init__(self, name=None, url=DEFAULT_CONTEXT_URL, enable_base=True, enable_testing=DEFAULT_ENABLE_TESTING, enable_source=False, enable_debuginfo=False):  #pylint: disable=too-many-arguments
        '''
            Setup the only really required attributes at init.

            :param name: name of the context
            :type name: str

            :param url: url of the context base
            :type url: str
            :default url: org_scientificlinux_contexts.constants.DEFAULT_CONTEXT_URL

            :param enable_base: Should the base repo be enabled?
            :type enable_base: boolean
            :default enable_base: True

            :param enable_testing: Should the testing repo be enabled?
            :type enable_testing: boolean
            :default enable_testing: org_scientificlinux_contexts.constants.DEFAULT_ENABLE_TESTING

            :param enable_source: Should the source repo be enabled?
            :type enable_source: boolean
            :default enable_source: False

            :param enable_debuginfo: Should the debuginfo repo be enabled?
            :type enable_debuginfo: boolean
            :default enable_debuginfo: False
        '''
        if not url.endswith('/'):
            url = url + '/'

        self.name = name
        self.url = url

        self.enable_base = enable_base
        self.enable_testing = bool(enable_testing)  # so that we don't alter out of scope
        self.enable_source = enable_source
        self.enable_debuginfo = enable_debuginfo

        self.alternate_urls = []

    def __str__(self):
        '''
            If only, alas you need org_scientificlinux_contexts.parser.makeline
        '''
        raise NotImplementedError('you need to use org_scientificlinux_contexts.parser.makeline')

    @property
    def gpgkey(self):
        '''
            The location of this file is fixed, but dependant on
             name and url, which are not
        '''
        return self.url + self.name + '/' + GPG_PREFIX + self.name

    @property
    def packages(self):
        '''
            Return content for adding to ksdata.packages
        '''
        packages = []
        packages.append(REPO_PACKAGE_PREFIX + self.name)
        packages.append('@' + self.name)
        return tuple(packages)

    @property
    def yumrepos(self):
        '''
            Build up a correctly constructed yum repo stanzas for all repos.
            Output should be written to a single file.

            :returns: a string that contains all yum repos for this context
            :rtype: str
        '''
        return self.base_yumrepo() + self.testing_yumrepo() + self.debuginfo_yumrepo() + self.source_yumrepo()

    @property
    def ksrepos(self):
        '''
            Build up a correctly constructed kickstart repo def for all enabled repos

            :returns: a string suitable for kickstart or None
            :rtype: str or None
        '''
        base = self.base_ksrepo()
        testing = self.testing_ksrepo(skip_if_unavailable=True)
        debuginfo = self.debuginfo_ksrepo(skip_if_unavailable=True)
        source = self.source_ksrepo(skip_if_unavailable=True)

        retval = ''
        if base:
            retval = base
        if testing:
            retval = retval + '\n' + testing
        if debuginfo:
            retval = retval + '\n' + debuginfo
        if source:
            retval = retval + '\n' + source

        if retval == '':
            retval = None
        return retval

    def write_yumrepos(self, repodir='/etc/yum.repos.d/'):
        '''
            Write out the yum repos into repodir

            :param repodir: Where to write the repo files
            :type repodir: str
            :default repodir: /etc/yum.repos.d/
        '''
        if not repodir.endswith('/'):
            repodir = repodir + '/'

        _fd = open(repodir + self.name + BASE_SUFFIX + '.repo', 'w', 1)
        _fd.write(self.base_yumrepo())
        _fd.close()

        _fd = open(repodir + self.name + TESTING_SUFFIX + '.repo', 'w', 1)
        _fd.write(self.testing_yumrepo(skip_if_unavailable=True))
        _fd.close()

        _fd = open(repodir + self.name + DEBUGINFO_SUFFIX + '.repo', 'w', 1)
        _fd.write(self.debuginfo_yumrepo(skip_if_unavailable=True))
        _fd.close()

        _fd = open(repodir + self.name + SRPMS_SUFFIX + '.repo', 'w', 1)
        _fd.write(self.source_yumrepo(skip_if_unavailable=True))
        _fd.close()


    def __yumrepo(self, reponame, baseurl, mirrorlist, enable, skip_if_unavailable):  # pylint: disable=unused-argument,too-many-arguments
        '''
            All the yum repos look almost alike, so isolate the differences there

            :param reponame: Name of the repo
            :type reponame: str

            :param baseurl: Url for the repo
            :type baseurl: str

            :param mirrorlist: Url for the mirror list
            :type mirrorlist: str

            :param enable: Should the repo be enabled
            :type enable: boolean

            :param skip_if_unavailable: Toggles the yum parm skip_if_unavailable
            :type skip_if_unavailable: boolean

            :returns: a string that contains the yum repo
            :rtype: str
        '''
        repostr = '\n'
        # Setup a basic repo
        repostr += '[%s]\n' % reponame
        repostr += 'name=%s\n' % reponame
        repostr += 'baseurl=%s\n' % baseurl
        repostr += 'mirrorlist=%s\n' % mirrorlist
        repostr += 'gpgcheck=1\n'
        repostr += 'gpgkey=%s\n' % self.gpgkey
        repostr += 'metadata_expire=60\n'
        if enable:
            repostr += 'enabled=1\n'
        else:
            repostr += 'enabled=0\n'
        if skip_if_unavailable:
            repostr += 'skip_if_unavailable=1\n'
        repostr += '\n'

        return repostr

    def __ksrepo(self, reponame, baseurl, mirrorlist, enable, skip_if_unavailable):  # pylint: disable=unused-argument,too-many-arguments
        '''
            All the ks repos look almost alike, so isolate the differences there

            :param reponame: Name of the repo
            :type reponame: str

            :param baseurl: Url for the repo
            :type baseurl: str

            :param mirrorlist: Url for the mirror list
            :type mirrorlist: str

            :param enable: Should the repo be enabled
            :type enable: boolean

            :param skip_if_unavailable: Toggles the yum parm skip_if_unavailable
            :type skip_if_unavailable: boolean

            :returns: a string that contains the yum repo
            :rtype: str
        '''
        if not enable:
            return None

        if mirrorlist:
            tmprepo = RepoData(name=reponame, mirrorlist=mirrorlist, enabled=True)
        else:
            tmprepo = RepoData(name=reponame, baseurl=baseurl, enabled=True)

        repostr = str(tmprepo)
        repostr = repostr.replace('"', '').replace("'", '')

        return repostr

    def base_yumrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed yum repo for base repo

            :param skip_if_unavailable: Toggles the yum parm skip_if_unavailable
            :type skip_if_unavailable: boolean
            :default skip_if_unavailable: org_scientificlinux_contexts.constants.YUMREPO_SKIP_IF_UNAVAILABLE

            :returns: a string that contains the yum repo
            :rtype: str
        '''
        baseurl = self.url + self.name + '/' + PATH_BASE
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_BASE
        reponame = self.name + BASE_SUFFIX

        repostr = self.__yumrepo(reponame, baseurl, mirrorlist, self.enable_base, skip_if_unavailable)

        return repostr

    def base_ksrepo(self):
        '''
            Build up a correctly constructed kickstart repo def for base repo

            :returns: a string suitable for kickstart or None
            :rtype: str or None
        '''
        baseurl = self.url + self.name + '/' + PATH_BASE
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_BASE
        reponame = self.name + BASE_SUFFIX

        repostr = self.__ksrepo(reponame, baseurl, mirrorlist, self.enable_base, False)

        return repostr

    def testing_yumrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed yum repo for testing repo

            :param skip_if_unavailable: Toggles the yum parm skip_if_unavailable
            :type skip_if_unavailable: boolean
            :default skip_if_unavailable: org_scientificlinux_contexts.constants.YUMREPO_SKIP_IF_UNAVAILABLE

            :returns: a string that contains the yum repo
            :rtype: str
        '''
        baseurl = self.url + self.name + '/' + PATH_TESTING
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_TESTING
        reponame = self.name + TESTING_SUFFIX

        repostr = self.__yumrepo(reponame, baseurl, mirrorlist, self.enable_testing, skip_if_unavailable)

        return repostr

    def testing_ksrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed kickstart repo def for testing repo

            :returns: a string suitable for kickstart or None
            :rtype: str or None
        '''
        baseurl = self.url + self.name + '/' + PATH_TESTING
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_TESTING
        reponame = self.name + TESTING_SUFFIX

        repostr = self.__ksrepo(reponame, baseurl, mirrorlist, self.enable_testing, skip_if_unavailable)

        return repostr

    def debuginfo_yumrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed yum repo for debuginfo repo

            :param skip_if_unavailable: Toggles the yum parm skip_if_unavailable
            :type skip_if_unavailable: boolean
            :default skip_if_unavailable: org_scientificlinux_contexts.constants.YUMREPO_SKIP_IF_UNAVAILABLE

            :returns: a string that contains the yum repo
            :rtype: str
        '''
        baseurl = self.url + self.name + '/' + PATH_DEBUGINFO
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_DEBUGINFO
        reponame = self.name + DEBUGINFO_SUFFIX

        repostr = self.__yumrepo(reponame, baseurl, mirrorlist, self.enable_debuginfo, skip_if_unavailable)

        return repostr

    def debuginfo_ksrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed kickstart repo def for debuginfo repo

            :returns: a string suitable for kickstart or None
            :rtype: str or None
        '''
        baseurl = self.url + self.name + '/' + PATH_DEBUGINFO
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_DEBUGINFO
        reponame = self.name + DEBUGINFO_SUFFIX

        repostr = self.__ksrepo(reponame, baseurl, mirrorlist, self.enable_debuginfo, skip_if_unavailable)

        return repostr

    def source_yumrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed yum repo for source repo

            :param skip_if_unavailable: Toggles the yum parm skip_if_unavailable
            :type skip_if_unavailable: boolean
            :default skip_if_unavailable: org_scientificlinux_contexts.constants.YUMREPO_SKIP_IF_UNAVAILABLE

            :returns: a string that contains the yum repo
            :rtype: str
        '''
        baseurl = self.url + self.name + '/' + PATH_SRPMS
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_SRPMS
        reponame = self.name + SRPMS_SUFFIX

        repostr = self.__yumrepo(reponame, baseurl, mirrorlist, self.enable_source, skip_if_unavailable)

        return repostr

    def source_ksrepo(self, skip_if_unavailable=YUMREPO_SKIP_IF_UNAVAILABLE):
        '''
            Build up a correctly constructed kickstart repo def for source repo

            :returns: a string suitable for kickstart or None
            :rtype: str or None
        '''
        baseurl = self.url + self.name + '/' + PATH_SRPMS
        mirrorlist = self.url + self.name + '/' + PATH_MIRROR_SRPMS
        reponame = self.name + SRPMS_SUFFIX

        repostr = self.__ksrepo(reponame, baseurl, mirrorlist, self.enable_source, skip_if_unavailable)

        return repostr

