#pylint: disable=line-too-long
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'''
    Functions universal to context behavior
'''
from __future__ import unicode_literals
from __future__ import absolute_import

import json
import logging
import os

from pyanaconda import constants
from pyanaconda import nm
from pyanaconda.i18n import _
from pyanaconda.threads import threadMgr, AnacondaThread

from org_scientificlinux_contexts.data_fetch import fetch_data
from org_scientificlinux_contexts.constants import ANACONDA_THREAD_FETCH_DATA, ADDON_NAME, DEFAULT_CONTEXT_URL, DEFAULT_CONTEXT_MD

log = logging.getLogger("anaconda") #pylint: disable=invalid-name

class addonError(Exception): #pylint: disable=invalid-name
    '''Just a place to gather exceptions'''
    pass
class addonErrorNetwork(addonError): #pylint: disable=invalid-name
    '''Network related error'''
    pass

def wait_for_network():
    '''
        Block until the network is live.

        :see: org_scientificlinux_contexts.constants.THREAD_WAIT_FOR_CONNECTING_NM
    '''
    # get thread that tries to establish a network connection
    nm_conn_thread = threadMgr.get(constants.THREAD_WAIT_FOR_CONNECTING_NM)
    if nm_conn_thread:
        log.debug("%s: NM still connecting, waiting for it to finish", ADDON_NAME)
        nm_conn_thread.join()

    if not nm.nm_is_connected():
        raise addonErrorNetwork("Network connection needed to fetch data.")

def fetch(url, out_file=None, ca_certs=None, proxy_server=None, proxy_port=3128, proxy_user=None, proxy_pass=None):
    '''
        Download target URL.
        If no out_file is provided, out_file is /tmp/ + os.path.basename(url)
        If you want a custom SSL CA cert, you can pass it in too

        :see: org_scientificlinux_contexts.data_fetch
        :param url: a valid url for a given resource
        :type url: string instance
        :param out_file: the location to write the file.
                         if None, defaults to /tmp/os.path.basename(url)
        :type out_file: string instance
        :param ca_certs: path to a CA cert bundle
        :type ca_certs: string instance
        :param proxy_server: address or FQDN of proxy server
        :type proxy_server: str
        :param proxy_port: port proxy server uses
        :type proxy_port: int
        :param proxy_user: username for proxy server
        :type proxy_user: str
        :param proxy_pass: password for proxy server
        :type proxy_pass: str
    '''
    if not out_file:
        out_file = '/tmp/' + os.path.basename(url)

    wait_for_network()

    fetch_data_thread = AnacondaThread(name=ANACONDA_THREAD_FETCH_DATA,
                                       target=fetch_data,
                                       args=(url, out_file, ca_certs, proxy_server, proxy_port, proxy_user, proxy_pass),
                                       fatal=False)

    log.debug("%s: Fetching: %s as %s", ADDON_NAME, url, out_file)
    # register and run the thread
    threadMgr.add(fetch_data_thread)

    return ANACONDA_THREAD_FETCH_DATA

def get_contexts(contexturl=DEFAULT_CONTEXT_URL, contextmd=DEFAULT_CONTEXT_MD, tmpdir='/run/org_scientificlinux_contexts/get_contexts/'):
    '''
        Return list of valid contexts at contexturl

        Processed files remain in tmpdir

        :see: org_scientificlinux_contexts.common.fetch

        :param contexturl: a valid url where the contextmd is kept
        :type url: string instance
        :default url: org_scientificlinux_contexts.constants.DEFAULT_CONTEXT_URL

        :param contextmd: name of the metadata file to fetch
        :type contextmd: string instance
        :default contextmd: org_scientificlinux_contexts.constants.DEFAULT_CONTEXT_MD

        :param tmpdir: The local path to store the retrieved metadata
        :type tmpdir: string instance
        :default tmpdir: /run/org_scientificlinux_contexts/get_contexts/

        :returns: a list of contexts at a given url
        :rtype: list
    '''
    contexts = []

    if contexturl.endswith('/'):
        url = contexturl + contextmd
    else:
        url = contexturl + '/' + contextmd

    outfile = url.replace('/', '_')

    if tmpdir.endswith('/'):
        outfile = tmpdir + outfile
    else:
        outfile = tmpdir + '/' + outfile

    log.debug("%s: Fetching context md from : %s", ADDON_NAME, url)

    if not os.path.isdir(tmpdir):
        os.makedirs(tmpdir, 0755)

    try:
        if not os.path.isfile(outfile):
            fetch(str(url), str(outfile))
            threadMgr.get(ANACONDA_THREAD_FETCH_DATA).join()
        else:
            log.debug("%s: already fetched %s", ADDON_NAME, outfile)

        if os.path.isfile(outfile):
            _fd = open(outfile, 'r')
            contexts = json.loads(_fd.read())
            _fd.close()
        else:
            log.warning("%s: Unable to fetch contexts JSON from %s", ADDON_NAME, url)
    except ValueError:
        log.warning("%s: Unable to read contexts JSON from %s", ADDON_NAME, url)

    log.debug("%s: Found contexts: %s at %s", ADDON_NAME, contexts, url)

    return contexts

def get_status(context_obj):
    '''
        Return a status message about our settings

        :param context_obj: our addon's kickstart data object
        :type context_obj: org_scientificlinux_contexts.ks.context.ContextKSData instance
    '''
    enabled = []
    for context in context_obj.enabled_contexts:
        if context_obj[context].enable_testing and not context_obj.testing:
            enabled.append(context + '(' + _("testing") + ')')
        else:
            enabled.append(context)

    if enabled:
        msg = _('Enabled Contexts')
        if context_obj.testing:
            msg = msg + '(' + _("testing") + ')'
        txt = ' '.join(enabled)
        msg = msg + ': ' + txt
        return msg
    elif not nm.nm_is_connected():
        return _('Network Not Connected')

    else:
        return _('None Enabled')

