#pylint: disable=line-too-long
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'''
    Location of the GUI specific behaviors
'''
# for python3 compat
from __future__ import unicode_literals
from __future__ import absolute_import

import logging
import os
import time
import urlparse

from threading import RLock

from pyanaconda import nm, safe_dbus

from pyanaconda.i18n import N_, _

from pyanaconda.flags import BootArgs

from pyanaconda.threads import threadMgr, AnacondaThread

from pyanaconda.ui.categories.software import SoftwareCategory
from pyanaconda.ui.communication import hubQ

from pyanaconda.ui.gui.spokes import NormalSpoke
from pyanaconda.ui.gui.utils import gtk_action_wait, really_hide, really_show

from org_scientificlinux_contexts import Context
from org_scientificlinux_contexts.common import get_contexts, get_status
from org_scientificlinux_contexts.constants import ADDON_NAME, DEFAULT_CONTEXT_URL, DEFAULT_ENABLE_TESTING
from org_scientificlinux_contexts.constants import DVD_CONTEXT_DIR, DEFAULT_CONTEXT_MD, BOOT_ARG_DEFAULTURL
from org_scientificlinux_contexts.constants import ANACONDA_THREAD_UI_FETCH_DATA
#, ANACONDA_THREAD_GUI_NM_CONNECTED

__all__ = ["ContextSpoke"]

log = logging.getLogger("anaconda") #pylint: disable=invalid-name

contextStore_LIST_Enabled = 0  #pylint: disable=invalid-name
contextStore_LIST_Testing = 1  #pylint: disable=invalid-name
contextStore_LIST_Name = 2     #pylint: disable=invalid-name
contextStore_LIST_Url = 3      #pylint: disable=invalid-name

def make_context_store_list(name, url=DEFAULT_CONTEXT_URL, enabled=False, testing=False):
    '''
        Glade is very sensitive to the order of elements within contextStore
        so this function makes a correctly constructed list

    '''
    mylist = [bool(enabled), bool(testing), name, url]
    return mylist

class FetchState(object): #pylint: disable=too-few-public-methods
    '''I'm just a simple object to make it easy to see if I'm fetching'''
    def __init__(self):
        '''I hold two values, a lock and a boolean status'''
        self.lock = RLock()
        self.isfetching = False
    def __enter__(self):
        ''' The python 'with' syntax runs __enter__ '''
        self.lock.acquire()
        self.isfetching = True
    def __exit__(self, _type, value, traceback):
        ''' The python 'with' syntax runs __exit__ '''
        self.isfetching = False
        self.lock.release()
    def __bool__(self):
        '''Am I fetching right now'''
        return self.isfetching
    def __nonzero__(self):
        '''Python2 compat function'''
        return type(self).__bool__(self)

class ContextSpoke(NormalSpoke): #pylint: disable=too-many-public-methods,too-many-ancestors,too-many-instance-attributes
    '''
        Main class of the Context addon spoke that will appear in the Software
        category on the Summary hub. It allows interactive choosing of the context
        its location, and possible testing instances.

        :see: pyanaconda.ui.common.UIObject
        :see: pyanaconda.ui.common.Spoke
        :see: pyanaconda.ui.gui.GUIObject
    '''
    # list all top-level objects from the .glade file that should be exposed
    # to the spoke or leave empty to extract everything
    builderObjects = ["contextSpokeWindow", "enabledContextStore", "knownContextStore"]

    # the name of the main window widget
    mainWidgetName = "contextSpokeWindow"

    # name of the .glade file in the same directory as this source
    uiFile = "context.glade"

    # category this spoke belongs to
    category = SoftwareCategory

    # spoke icon (will be displayed on the hub)
    # preferred are the -symbolic icons as these are used in Anaconda's spokes
    ###icon = "insert-object-symbolic"
    icon = "preferences-other-symbolic"

    # title of the spoke (will be displayed on the hub)
    title = N_("CONTEXT")

    ### methods defined by API and helper methods ###
    def __init__(self, data, storage, payload, instclass):
        """
            :see: pyanaconda.ui.common.Spoke.__init__
            :param data: data object passed to every spoke to load/store data
                         from/to it
            :type data: pykickstart.base.BaseHandler
            :param storage: object storing storage-related information
                            (disks, partitioning, bootloader, etc.)
            :type storage: blivet.Blivet
            :param payload: object storing packaging-related information
            :type payload: pyanaconda.packaging.Payload
            :param instclass: distribution-specific information
            :type instclass: pyanaconda.installclass.BaseInstallClass
        """
        NormalSpoke.__init__(self, data, storage, payload, instclass)

        # self.args holds the master data set
        try:
            self.args = self.data.addons.org_scientificlinux_contexts
        except AttributeError:
            log.debug("%s: couldn't add args, unit tests running?", ADDON_NAME)

        try:
            self.name = self.data.addons.org_scientificlinux_contexts.name
        except AttributeError:
            self.name = ADDON_NAME
            log.debug("%s: addon name fall through, unit tests running?", self.name)

        try:
            self.args.ksdata = data
            log.debug("%s: adding ksdata to kickstart object", self.name)
        except AttributeError:
            log.debug("%s: couldn't add ksdata, unit tests running?", self.name)

        # allow commandline override
        self.scanurl = DEFAULT_CONTEXT_URL
        cmdline = BootArgs()
        if BOOT_ARG_DEFAULTURL in cmdline:
            self.scanurl = cmdline[BOOT_ARG_DEFAULTURL]
        log.debug("%s: setting default scanurl: %s", self.name, self.scanurl)

        self.__fetching = FetchState()
        self.__dbus_con = safe_dbus.get_new_session_connection()

        # These get defined correctly within initalize
        #pylint: disable=invalid-name
        self.__urlEntry = None
        self.__scanURLButton = None

        self.__knownContextStore = None
        self.__enabledContextStore = None

        self.__contextGlobalTestingSwitch = None

        self.__ready = False

        self.__fetchingLabel = None
        self.__fetchingSpinner = None

    def initialize(self):
        """
            The initialize method that is called after the instance is created.
            The difference between __init__ and this method is that this may take
            a long time and thus could be called in a separated thread.

            :see: pyanaconda.ui.common.UIObject.initialize
        """
        log.debug("%s: gui initialize", self.name)
        NormalSpoke.initialize(self)

        self.__knownContextStore = self.builder.get_object("knownContextStore")
        self.__enabledContextStore = self.builder.get_object("enabledContextStore")

        self.__contextGlobalTestingSwitch = self.builder.get_object("contextGlobalTestingSwitch")

        self.__urlEntry = self.builder.get_object("urlEntry")
        self.__scanURLButton = self.builder.get_object("scanURLButton")

        self.__fetchingLabel = self.builder.get_object("fetchingLabel")
        self.__fetchingSpinner = self.builder.get_object("fetchingSpinner")

        if self.scanurl.startswith('file:'):
            self.fetch_contexts(self.scanurl)
        if os.path.isfile(DVD_CONTEXT_DIR + '/' + DEFAULT_CONTEXT_MD):
            self.fetch_contexts('file://' + DVD_CONTEXT_DIR + '/' + DEFAULT_CONTEXT_MD)

        self.__ready = True

        # dbus watches for any changes to NetworkManager and fires off this function
        self.__dbus_con.signal_subscribe('org.freedesktop.NetworkManager', 'org.freedesktop.NetworkManager.Settings.Connection', None, None, None, 0, self.__ui_watch_for_network)

    def __ui_watch_for_network(self):
        '''
            When the network appears, send an update to the UI
        '''
        hubQ.send_message(self.__class__.__name__, get_status(self.args))  #pylint: disable=no-member

        log.debug("%s: network state check ran", self.name)

    def fetch_contexts(self, scanurl=None):
        '''
            Spawns a thread that will fetch context data, update the known contexts
            and refresh the UI
        '''
        if scanurl == None:
            scanurl = self.scanurl
        if nm.nm_is_connected() or scanurl.startswith('file:'):
            log.debug("%s: gui fetch_contexts: %s", self.name, scanurl)

            # These can fire off too quickly....
            if threadMgr.exists(ANACONDA_THREAD_UI_FETCH_DATA):
                threadMgr.wait(ANACONDA_THREAD_UI_FETCH_DATA)

            fetch_data_thread = AnacondaThread(name=ANACONDA_THREAD_UI_FETCH_DATA,
                                               target=self.__background_fetching, fatal=True)
            # register and run the thread
            threadMgr.add(fetch_data_thread)
        else:
            log.debug("%s: gui fetch_contexts no network for: %s", self.name, scanurl)
        return ANACONDA_THREAD_UI_FETCH_DATA

    def __background_fetching(self):
        '''
            The actual code for the fetching thread

            :param url: What to fetch
            :type url: str
        '''
        log.debug("%s: gui __background_fetching waiting: %s", self.name, self.scanurl)
        with self.__fetching:
            log.debug("%s: gui __background_fetching start: %s", self.name, self.scanurl)
            for name in get_contexts(self.scanurl):
                if name in self.args:
                    if self.scanurl == self.args[name].url:
                        continue
                    self.args[name].alternate_urls.append(self.scanurl)
                else:
                    newcontext = Context(name=name, url=self.scanurl, enable_base=False, enable_testing=DEFAULT_ENABLE_TESTING)
                    self.args.add(newcontext)
            log.debug("%s: gui __background_fetching end: %s", self.name, self.scanurl)

        self.refresh()

    @gtk_action_wait
    def refresh(self):
        """
            The refresh method that is called every time the spoke is displayed.
            It should update the UI elements according to the contents of
            self.data.

            :see: pyanaconda.ui.common.UIObject.refresh
        """
        log.debug("%s: gui refresh", self.name)

        if not self.args.known_contexts:
            log.debug("%s: gui refresh no known contexts, fetching default", self.name)
            self.fetch_contexts()

        self.__urlEntry.set_text(self.scanurl)

        if self.args.testing:
            self.__contextGlobalTestingSwitch.set_active(True)
        else:
            self.__contextGlobalTestingSwitch.set_active(False)

        self.__enabledContextStore.clear()
        self.__knownContextStore.clear()
        # add default location/enabled ones at the top
        for context in self.args.known_contexts:
            myattribs = make_context_store_list(url=self.args[context].url,
                                                name=self.args[context].name,
                                                enabled=self.args[context].enable_base,
                                                testing=self.args[context].enable_testing)
            if self.args[context].enable_base:
                self.__enabledContextStore.append(myattribs)
            else:
                self.__knownContextStore.append(myattribs)

            alt_loc = self.args[context].alternate_urls
            alt_loc.sort()
            for url in alt_loc:
                # testing tracks state across all locations
                myattribs = make_context_store_list(url=url,
                                                    name=self.args[context].name,
                                                    enabled=False,
                                                    testing=self.args[context].enable_testing)
                self.__knownContextStore.append(myattribs)

        if self.__fetching:
            really_show(self.__fetchingLabel)
            really_show(self.__fetchingSpinner)
            really_hide(self.__scanURLButton)
            self.__urlEntry.set_sensitive(False)
        else:
            really_hide(self.__fetchingLabel)
            really_hide(self.__fetchingSpinner)
            really_show(self.__scanURLButton)
            self.__urlEntry.set_sensitive(True)

        if not nm.nm_is_connected():
            # Basically, the network isn't up so disable everything
            really_hide(self.__fetchingLabel)
            really_hide(self.__fetchingSpinner)
            really_hide(self.__scanURLButton)
            self.__contextGlobalTestingSwitch.set_sensitive(False)
            self.__urlEntry.set_sensitive(False)
            self.__urlEntry.set_text(get_status(self.args))
            self.set_info(_("Network connection needed to fetch data."))
        else:
            self.__contextGlobalTestingSwitch.set_sensitive(True)
            self.clear_info()

    def apply(self):
        """
            The apply method that is called when the spoke is left. It should
            update the contents of self.data with values set in the GUI elements.
        """
        # I'm functionally doing this async with the GTK widgets
        pass

    def execute(self):
        """
            The excecute method that is called when the spoke is left. It is
            supposed to do all changes to the runtime environment according to
            the values set in the GUI elements.
        """
        # I'm functionally doing this async with the GTK widgets
        pass

    @property
    def ready(self):
        """
            The ready property that tells whether the spoke is ready
            (can be visited) or not.

            :rtype: bool
        """
        # basically, this is not ready until initialize is done
        # TODO: find ready state of software and sources
        #       use that as a baseline for the readyness here
        return self.__ready

    @property
    def completed(self):
        """
            The completed property that tells whether all mandatory items on the
            spoke are set, or not. The spoke will be marked on the hub as completed
            or uncompleted acording to the returned value.

            :rtype: bool
        """
        # this is an optional spoke and is not mandatory, so sure we are 'done'
        # TODO: figure out the "after install initial setup" ini config
        #       may relate to the 'apply' bits....
        return True

    @property
    def mandatory(self):
        """
            The mandatory property that tells whether the spoke is mandatory to be
            completed to continue in the installation process.

            :rtype: bool
        """
        # this is an optional spoke and is not mandatory
        return False

    @property
    @gtk_action_wait
    def status(self):
        """
            The status property that is a brief string describing the state of the
            spoke. It should describe whether all values are set and if possible
            also the values themselves. The returned value will appear on the hub
            below the spoke's title.

            :rtype: str
        """
        return get_status(self.args)

    def on_scanURLButton_clicked(self, url): #pylint: disable=invalid-name
        """ Called when the Scan URL button is clicked """
        log.debug("%s: on_scanURLButton_clicked: clicked %s", self.name, url.get_text())

        self.scanurl = str(url.get_text())

        result = urlparse.urlparse(self.scanurl)
        if not result.scheme:
            self.scanurl = 'INVALID URL'
        else:
            self.fetch_contexts(self.scanurl)

        self.refresh()

    def on_enabledContextEnable_toggled(self, renderer, path):  # pylint: disable=unused-argument,invalid-name
        """ Called when the context Enable checkbox is clicked """
        enabled = not bool(self.__enabledContextStore[path][contextStore_LIST_Enabled])
        name = self.__enabledContextStore[path][contextStore_LIST_Name]
        url = self.__enabledContextStore[path][contextStore_LIST_Url]

        self.args[name].enable_base = enabled
        if url != self.args[name].url:
            self.args[name].alternate_urls.append(self.args[name].url)
            self.args[name].alternate_urls.remove(url)
            self.args[name].url = url

        log.debug("%s: on_enabledContextEnable_toggled: set %s(%s) enabled %s", self.name, url, name, enabled)
        self.refresh()

    def on_enabledContextEnableTesting_toggled(self, renderer, path):  # pylint: disable=unused-argument,invalid-name
        """ Called when the context Enable checkbox is clicked """
        enabled = not bool(self.__enabledContextStore[path][contextStore_LIST_Testing])
        name = self.__enabledContextStore[path][contextStore_LIST_Name]

        self.args[name].enable_testing = enabled

        log.debug("%s: on_enabledContextEnableTesting_toggled: set %s testing enabled %s", self.name, name, enabled)
        self.refresh()

    def on_knownContextEnableRenderer_toggled(self, renderer, path):  # pylint: disable=unused-argument,invalid-name
        """ Called when the context Enable checkbox is clicked """
        enabled = not bool(self.__knownContextStore[path][contextStore_LIST_Enabled])
        name = self.__knownContextStore[path][contextStore_LIST_Name]
        url = self.__knownContextStore[path][contextStore_LIST_Url]

        self.args[name].enable_base = enabled
        if url != self.args[name].url:
            self.args[name].alternate_urls.append(self.args[name].url)
            self.args[name].alternate_urls.remove(url)
            self.args[name].url = url

        log.debug("%s: on_knownContextEnableRenderer_toggled: set %s(%s) enabled %s", self.name, url, name, enabled)
        self.refresh()

    def on_contextGlobalTestingSwitch_toggled(self, switch, *args):  # pylint: disable=unused-argument,invalid-name
        '''
            Make the change in the real object when this changes
        '''
        self.args.testing = switch.get_active()
        log.debug("%s: on_contextGlobalTestingSwitch_toggled: set %s", self.name, self.args.testing)
        self.refresh()

