#pylint: disable=line-too-long
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'''
    Location of the kickstart specific behaviors for
    org_scientificlinux_contexts
'''
# for python3 compat
from __future__ import unicode_literals
from __future__ import absolute_import

import ConfigParser
import logging
import os
import re
import shutil
import subprocess

from pyanaconda import iutil
from pyanaconda.i18n import _
from pyanaconda.addons import AddonData
from pyanaconda.flags import BootArgs
from pyanaconda.progress import progressQ

from pykickstart.options import KSOptionParser
from pykickstart.errors import KickstartParseError, KickstartValueError, formatErrorMsg

from org_scientificlinux_contexts import parser
from org_scientificlinux_contexts.version import __version__
from org_scientificlinux_contexts.common import get_contexts
from org_scientificlinux_contexts.constants import CHROOT_BASE, DEFAULT_ENABLE_TESTING, DEFAULT_CONTEXT_URL
from org_scientificlinux_contexts.constants import BOOT_ARG_ENABLE, BOOT_ARG_TESTING, BOOT_ARG_DEFAULTURL
from org_scientificlinux_contexts.constants import ANACONDA_YUM_CONF, ANACONDA_YUM_REPOS, DVD_MOUNTED_ON


# export ContextKSData class to prevent Anaconda's collect method from taking
# AddonData class instead of the ContextKSData class
# :see: pyanaconda.kickstart.AnacondaKSHandler.__init__
__all__ = ["ContextKSData"]

log = logging.getLogger("anaconda") #pylint: disable=invalid-name

class ContextKSData(AddonData):
    '''
        Class parsing and storing data for the context addon (kickstart)
    '''
    def __init__(self, name):
        """
            :param name: name of the addon
            :type name: str

        """
        AddonData.__init__(self, name)
        log.debug("%s: running __init__ for kickstart version:%s", self.name, __version__)

        self.testing = bool(DEFAULT_ENABLE_TESTING) # so that we don't alter the constant
        self.defaulturl = str(DEFAULT_CONTEXT_URL)  # so that we don't alter the constant

        self.__contexts = {}

        self.__read_proc_cmdline()

        self.__from_dvd = os.path.ismount(DVD_MOUNTED_ON)

        self.__mountpoints = {}

        with open('/proc/mounts', 'r') as _fd:
            for line in _fd.readlines():
                fields = line.split()
                self.__mountpoints[fields[1]] = fields[0]
                log.debug("%s: mountpoint %s -> %s", self.name, fields[0], fields[1])

        points = self.__mountpoints.keys()
        points.sort()
        log.debug("%s: mountpoints: %s", self.name, points)

        self.ksdata = None

    def __str__(self):
        """
            What should end up in the resulting kickstart file, i.e. the %addon
            section containing string representation of the stored data.

            :returns: The text we paste into the kickstart file
            :rtype: str
        """
        log.debug("%s: running __str__ for kickstart", self.name)

        if not self.enabled_contexts:
            log.debug("%s: no contexts enabled", self.name)
            return"%%addon %s\n%%end\n\n" % self.name

        addon_str = '# Scientific Linux Contexts\n'

        addon_str += "%%addon %s " % self.name
        if self.testing:
            addon_str += "--testing"
        if self.defaulturl != DEFAULT_CONTEXT_URL:
            addon_str += " --defaulturl=%s" %self.defaulturl
        addon_str += "\n"

        for context in self.enabled_contexts:
            addon_str += parser.makeline(self[context], default_url=self.defaulturl, default_testing=self.testing)

        addon_str += "%end\n\n"

        for line in addon_str.split('\n'):
            log.debug("%s: __str__ : %s", self.name, line)

        return addon_str

    def __getitem__(self, key):
        '''
            Used for: treating like a dict

            :param key: the name of a context we know about
            :type key: string instance

            :returns: the context you requested
            :rtype: org_scientificlinux_contexts.Context
        '''
        try:
            return self.__contexts[key]
        except KeyError:
            return self.__contexts[key.name]

    def __setitem__(self, key, item):
        '''
            Used for: treating like a dict

            :param key: the name of a context
            :type key: string instance

            :param item: the context object
            :type item: org_scientificlinux_contexts.Context
        '''
        if key != item.name:
            raise ValueError('key:' + key + ' != ' + item.name)
        self.__contexts[key] = item

    def __delitem__(self, key):
        '''
            Used for: treating like a dict

            :param key: the name of a context we know about to remove
            :type key: string instance
        '''
        try:
            del self.__contexts[key]
        except KeyError:
            del self.__contexts[key.name]

    def __contains__(self, key):
        '''
            Used for: x in object

            :param key: the name of a context to see if we know it exists
            :type key: string instance
        '''
        if key in self.keys():
            return True
        try:
            if key.name in self.keys():
                return True
        except AttributeError:
            pass

        return False

    def keys(self):
        '''
            Used for: treating like a dict

            :returns: a list of known contexts
            :rtype: list
        '''
        retval = self.__contexts.keys()
        retval.sort()

        return retval

    @property
    def known_contexts(self):
        '''
            List out the names of all known contexts

            :returns: a list of known contexts
            :rtype: list
        '''
        return self.keys()

    @property
    def enabled_contexts(self):
        '''
            List out the names of all enabled contexts

            :returns: a list of enabled contexts
            :rtype: list
        '''
        retval = []
        for context in self.known_contexts:
            if self[context].enable_base:
                retval.append(context)

        retval.sort()
        return retval

    def add(self, context):
        '''
            Adds a context to the list of enabled contexts

            :param context: the context to enable
            :type context: org_scientificlinux_contexts.Context
        '''
        self[context.name] = context

    def remove(self, context):
        '''
            Removes a context by name from the list of enabled contexts

            :param context: the context to remove
            :type context: string instance
        '''
        del self[context]

    def __read_proc_cmdline(self):
        """
            Read any args that were passed in via /proc/cmdline.
        """
        log.debug("%s: running __read_proc_cmdline for kickstart", self.name)

        cmdline = BootArgs()

        if BOOT_ARG_TESTING in cmdline:
            log.debug("%s: running __read_proc_cmdline for kickstart: setting testing == True", self.name)
            self.testing = True

        if BOOT_ARG_DEFAULTURL in cmdline:
            log.debug("%s: running __read_proc_cmdline for kickstart: setting defaulturl == %s", self.name, cmdline[BOOT_ARG_DEFAULTURL])
            self.defaulturl = cmdline[BOOT_ARG_DEFAULTURL]

        if BOOT_ARG_ENABLE in cmdline:
            for mytext in cmdline[BOOT_ARG_ENABLE].split(','):
                log.debug("%s: running __read_proc_cmdline parse for kickstart: reading %s", self.name, mytext)
                result = parser.readline(mytext)

                if result:
                    log.debug("%s: running __read_proc_cmdline parse for kickstart: added %s", self.name, result.name)
                    if result.enable_testing:
                        log.debug("%s: running __read_proc_cmdline parse for kickstart: enabled testing %s", self.name, result.name)
                    self.add(result)

    def handle_header(self, lineno, args):
        """
            Process additional arguments to the %addon line.

            This function receives any arguments on the %addon line after the
            addon ID. For example, for the line:

                %addon com_example_foo --argument='example'

            This function would be called with args=["--argument='example'"].

            By default AddonData.handle_header just preserves the passed
            arguments by storing them and adding them to the __str__ output.

            :param lineno: the current line number in the kickstart file
            :type lineno: int

            :param args: the list of arguments from the %addon line
            :type args: list
        """
        log.debug("%s: running handle_header for kickstart", self.name)
        _op = KSOptionParser()
        _op.add_option('--testing', action='store_true',
                       default=DEFAULT_ENABLE_TESTING,
                       help="enable the testing repo for your contexts")
        _op.add_option('--defaulturl',
                       default=DEFAULT_CONTEXT_URL,
                       help="set an alternate default url for contexts")

        (opts, extra) = _op.parse_args(args=args, lineno=lineno)

        log.debug("%s: argparse %s", self.name, opts)

        if opts.testing:
            self.testing = True
            log.debug("%s: argparse: set testing = True", self.name)
        if opts.defaulturl != DEFAULT_CONTEXT_URL:
            self.defaulturl = opts.defaulturl
            log.debug("%s: argparse: set defaulturl = %s", self.name, opts.defaulturl)

        # Reject any additional arguments.
        if extra:
            msg = "%s: Unhandled arguments on %%addon line" % self.name
            if lineno != None:
                raise KickstartParseError(formatErrorMsg(lineno, msg=msg))
            else:
                raise KickstartParseError(msg)

    def handle_line(self, line):
        """
            The handle_line method that is called with every line from
            this addon's %addon section of the kickstart file.

            :param line: a single line from the %addon section
            :type line: str

        """
        log.debug("%s: running handle_line for kickstart", self.name)
        # white space is ignored.
        mytext = line.lstrip().rstrip()

        log.debug("%s: running handle_line for kickstart: reading %s", self.name, mytext)
        result = parser.readline(mytext)

        if result:
            if result.name in self.known_contexts:
                log.warning("%s: handle_line for kickstart: already have %s", self.name, result.name)
            else:
                log.debug("%s: handle_line for kickstart: added %s", self.name, result.name)
                if result.enable_testing:
                    log.debug("%s: handle_line for kickstart: enabled testing %s", self.name, result.name)
                self.add(result)

    def setup(self, storage, ksdata, instclass, payload):  # pylint: disable=unused-argument
        """
            Make the changes to the install system.

            This method is called before the installation
            is started and directly from spokes. It must be possible
            to call it multiple times without breaking the environment.

            :param storage: object storing storage-related information
                            (disks, partitioning, bootloader, etc.)
            :type storage: blivet.Blivet instance
            :param ksdata: data parsed from the kickstart file and set in the
                           installation process
            :type ksdata: pykickstart.base.BaseHandler instance
            :param instclass: distribution-specific information
            :type instclass: pyanaconda.installclass.BaseInstallClass
            :param payload: object managing packages and environment groups
                            for the installation
        """
        log.debug("%s: running setup for kickstart", self.name)

        if self.enabled_contexts:
            if 'yum' not in ksdata.packages.packageList:
                log.debug("%s: adding yum to installed system", self.name)
                ksdata.packages.add(['yum'])
            else:
                log.debug("%s: yum already tagged for system this", self.name)

        return True

    def finalize(self):
        """
            The finalize method that is called when the end of the
            %addon section (i.e. when the %end line is processed).

            An addon should check if it has all required data.
            If not, it may handle the case quietly or it may raise
            the KickstartValueError exception.

            No additional data will come.

            Addon should check if all mandatory attributes were populated.

            :see: org_scientificlinux_contexts.common.get_contexts
        """
        log.debug("%s: running finalize for kickstart", self.name)

        if self.enabled_contexts and self.ksdata:
            if 'yum' not in self.ksdata.packages.packageList:
                log.debug("%s: adding yum to installed system", self.name)
                self.ksdata.packages.add(['yum'])
            else:
                log.debug("%s: yum already tagged for system this", self.name)

        # Make sure if we asked for a context there is somewhere to look for it
        urls = {}
        for context in self.enabled_contexts:
            if self[context].url not in urls:
                urls[self[context].url] = []
            urls[self[context].url].append(self[context].name)

        for url in urls:
            known_contexts = get_contexts(url)
            for context in urls[url]:
                if context not in known_contexts:
                    msg = "context:%s not found in %s at %s" % (context, known_contexts, url)
                    raise KickstartValueError(msg)

        if self.testing:
            for context in self.enabled_contexts:
                self[context].enable_testing = True
                log.debug("%s: force enable testing context for %s", self.name, self[context].name)

    def execute(self, storage, ksdata, instclass, users, payload): # pylint: disable=unused-argument
        """
            Make the changes to the underlying system.

            The execute method that should make changes to the installed system.
            It is called only once in the post-install setup phase.

            :see: setup

            :param storage: object storing storage-related information
                            (disks, partitioning, bootloader, etc.)
            :type storage: blivet.Blivet

            :param ksdata: data object passed to every spoke to load/store data
                           from/to it
            :type ksdata: pykickstart.base.BaseHandler

            :param instclass: distribution-specific information
            :type instclass: pyanaconda.installclass.BaseInstallClass

            :param users: information about created users
            :type users: pyanaconda.users.Users instance

            :param payload: object managing packages and environment groups
                            for the installation
        """
        log.debug("%s: running execute for kickstart", self.name)
        if not self.enabled_contexts:
            # nothing to do, return now
            log.debug("%s: no contexts enabled", self.name)
            return True

        self.__mount_dvd()

        # We MUST have yum installed in the new system
        ksdata.packages.add(['yum'])

        msg = _("Configuring")
        progressQ.send_message(msg + " Scientific Linux Context Framework") #pylint: disable=no-member

        # these files must be accessible within the chroot
        # /run is mounted in both places so it is easy to review if things go wrong
        repodir = '/run/org_scientificlinux_contexts/repos.d'
        yumconf_file = '/run/org_scientificlinux_contexts/yum.conf'

        self.__setup_yumrepos(ksdata, repodir, yumconf_file)

        # Clean up the metadata and any weird caches
        for output in iutil.execReadlines("yum", ['-c', yumconf_file, '--enablerepo=*', 'clean', 'all'], root=iutil.getSysroot()):
            log.debug("%s: (pre) yum-clean-all: %s", self.name, output)

        for context in self.enabled_contexts:
            log.debug("%s: loading context: %s", self.name, self[context].name)
            msg = _("Loading") + " Scientific Linux Context - " + self[context].name
            if self[context].enable_testing:
                msg = msg + ' ' + _("testing")
            progressQ.send_message(msg) #pylint: disable=no-member

            yum_args = ['-c', yumconf_file, '-y', 'install']
            yum_args += self[context].packages

            found_transaction = False
            for output in iutil.execReadlines("yum", yum_args, root=iutil.getSysroot()):
                log.debug("%s: yum load %s: %s", self.name, self[context].name, output)
                found_transaction = self.__show_transaction(msg, output, found_transaction)

        msg = _("Finalizing") + " Scientific Linux Contexts"
        progressQ.send_message(msg) #pylint: disable=no-member
        for output in iutil.execReadlines("yum", ['-c', yumconf_file, '-y', 'group', 'update'], root=iutil.getSysroot()):
            log.debug("%s: yum group update: %s", self.name, output)
            self.__show_transaction(msg, output)

        # Now that we are done, make sure to clean up after ourselves
        for output in iutil.execReadlines("yum", ['-c', yumconf_file, '--enablerepo=*', 'clean', 'all'], root=iutil.getSysroot()):
            log.debug("%s: (post) yum-clean-all: %s", self.name, output)

    @staticmethod
    def __show_transaction(prefix, line, found=False):
        '''
            Parse this output line and see if there is anything to show

            Ultimately this keys off of specific lines in the output
            so we use found to retain our state from line to line.

            :param prefix: the text to prefix the what we show
            :type prefix: string instance

            :param line: the line to parse out
            :type line: string instance
        '''
        if found:
            for action in ('Installing', 'Updating', 'Verifying', 'Cleanup'):
                if line.startswith(' ' + action) or line.startswith(action):
                    line_elements = re.split(r'\s+', line)
                    txt = ' '.join(line_elements)
                    progressQ.send_message(prefix + ' | ' + txt) #pylint: disable=no-member
                elif 'time: ' in line:
                    # raw parsing the yum log.... its ugly
                    found = False
                    progressQ.send_message(prefix) #pylint: disable=no-member
        elif line == 'Running transaction':
            # raw parsing the yum log.... its ugly
            found = True

        return found

    def __mount_dvd(self):
        '''
            Mount the DVD so non-network installs work right
        '''
        try:
            os.makedirs(DVD_MOUNTED_ON)
            os.makedirs(CHROOT_BASE + '/' + DVD_MOUNTED_ON)
        except OSError:
            pass

        if not os.path.ismount(DVD_MOUNTED_ON):
            if DVD_MOUNTED_ON in self.__mountpoints:
                # this was mounted earlier, so make sure it is mounted
                subprocess.check_call(['mount', self.__mountpoints[DVD_MOUNTED_ON], DVD_MOUNTED_ON], shell=True)
                log.debug("%s: mounted %s on %s", self.name, self.__mountpoints[DVD_MOUNTED_ON], DVD_MOUNTED_ON)
            else:
                subprocess.call(['mount', '/dev/sr0', DVD_MOUNTED_ON], shell=True)

        if not os.path.ismount(CHROOT_BASE + '/' + DVD_MOUNTED_ON):
            if DVD_MOUNTED_ON in self.__mountpoints:
                # this was mounted earlier, so make sure it is mounted in chroot
                iutil.execInSysroot('mount', [self.__mountpoints[DVD_MOUNTED_ON], DVD_MOUNTED_ON])
                log.debug("%s: mounted %s on %s", self.name, self.__mountpoints[DVD_MOUNTED_ON], CHROOT_BASE + '/' + DVD_MOUNTED_ON)
            else:
                iutil.execInSysroot('mount', ['/dev/sr0', DVD_MOUNTED_ON])


    def __setup_yumrepos(self, ksdata, repodir, yumconf_file):
        '''
            Copy in the repos for use within the chroot

            :param ksdata: data object passed to every spoke to load/store data
                           from/to it
            :type ksdata: pykickstart.base.BaseHandler

            :param repodir: where should the repos go
            :type repodir: string instance

            :param yumconf_file: where to write the yum.conf we are building
            :type yumconf_file: string instance
        '''
        # copy over any kickstart repos
        shutil.copytree(ANACONDA_YUM_REPOS, iutil.getSysroot() + repodir)

        # build yum config from the one we used to install the system
        yumconfig = self.__setup_yumconfig(ANACONDA_YUM_CONF, repodir)
        with open(yumconf_file, 'w') as _fd:
            yumconfig.write(_fd)

        with open(yumconf_file, 'r') as _fd:
            for line in _fd:
                log.debug("%s: %s: %s", self.name, yumconf_file, line)
            log.debug("%s: %s: #end of file", self.name, yumconf_file)

        log.debug("%s: %s contains: %s", self.name, repodir, os.listdir(repodir))

        for filename in os.listdir(repodir):
            with open(repodir + '/' + filename, 'r') as _fd:
                for line in _fd:
                    log.debug("%s: %s: %s", self.name, filename, line)
                log.debug("%s: %s: #end of file", self.name, filename)

        for context in self.enabled_contexts:
            log.debug("%s: enabled context %s at %s", self.name, self[context].name, self[context].url)
            if self[context].enable_testing:
                log.debug("%s: enabled TESTING context %s at %s", self.name, self[context].name, self[context].url)
            # while we are here, write out the repo files
            self[context].write_yumrepos(iutil.getSysroot() + repodir)

            # now update the kickstart file for non context enabled anacondas
            self.__add_repo_and_packages_to_ks(context, ksdata)

    def __add_repo_and_packages_to_ks(self, context, ksdata):
        '''
            Make changes to the kickstart file to support non-context
            anacondas

            :param context: name of the context to add
            :type context: string instance

            :param ksdata: information about the install class
            :type ksdata: pykickstart instance
        '''
        if self[context].ksrepos:
            log.debug("%s: all repos for non context anaconda:\n%s", self.name, self[context].ksrepos)
            for line in self[context].ksrepos.split('\n'):
                if not line:
                    # must be a blank line we inserted for clarity, ignore it
                    continue
                repostr = line.split(' ')
                log.debug("%s: adding repo for non context anaconda: %s", self.name, repostr)
                ksdata.dispatcher(repostr, lineno=1)
        ksdata.packages.add(self[context].packages)

    @staticmethod
    def __setup_yumconfig(filename=ANACONDA_YUM_CONF, repobase='/run/org_scientificlinux_contexts/repos.d'):
        '''
            Our execution of yum must ensure that we can find the context
            yum repos, so we must generate a viable yum conf that works with
            the existing install environment and contains our yum repos.

            :param filename: the location of anaconda's generated yum.conf
            :type filename: string instance
            :type default: org_scientificlinux_contexts.constants.ANACONDA_YUM_CONF

            :param repobase: the location of the yum repos we plan to parse
            :type repobase: string instance
            :type default: /run/org_scientificlinux_contexts/repos.d
        '''
        # make sure the yum-conf can be found for installation
        yumconfig = ConfigParser.SafeConfigParser()
        if os.path.isfile(filename):
            yumconfig.read(filename)
        else:
            yumconfig.add_section('main')

        # Make sure the metadata can be refreshed
        # Feel free to use the normal yum cache
        yumconfig.set('main', 'metadata_expire', '1')
        yumconfig.set('main', 'plugins', '1')
        yumconfig.set('main', 'debuglevel', '3')
        yumconfig.set('main', 'errorlevel', '5')

        # don't crash if some repos can't be found
        #  I'm looking at you /run/install/repo/ (DVD install off of local media)
        yumconfig.set('main', 'skip_if_unavailable', 'true')

        # Use my yum repos from the installer
        yumconfig.set('main', 'reposdir', repobase)

        # use default yum.log
        if yumconfig.has_option('main', 'logfile'):
            yumconfig.remove_option('main', 'logfile')
        # use default cache
        if yumconfig.has_option('main', 'cachedir'):
            yumconfig.remove_option('main', 'cachedir')

        return yumconfig
