#pylint: disable=line-too-long
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'''
    Location of the TUI specific behaviors
'''
# for python3 compat
from __future__ import unicode_literals
from __future__ import absolute_import

import logging
import re

from pyanaconda.i18n import N_, _

from pyanaconda.ui.categories.software import SoftwareCategory

from pyanaconda.constants_text import INPUT_PROCESSED

from pyanaconda.ui.tui.simpleline import TextWidget, ColumnWidget
from pyanaconda.ui.tui.spokes import EditTUISpoke
from pyanaconda.ui.tui.spokes import EditTUISpokeEntry as Entry  #pylint: disable=abstract-class-not-used

from org_scientificlinux_contexts import Context
from org_scientificlinux_contexts.common import get_status

__all__ = ["ContextSpoke"]

log = logging.getLogger("anaconda") #pylint: disable=invalid-name

class ContextSpoke(EditTUISpoke): #pylint: disable=too-many-public-methods,too-many-ancestors
    ''' The Context top level TUI interface '''
    title = N_("Context")
    category = SoftwareCategory

    def __init__(self, app, data, storage, payload, instclass): #pylint: disable=too-many-arguments
        '''
            self.args should hold the object this Spoke is supposed to edit

            :see: pyanaconda.ui.common.Spoke.__init__
            :param data: data object passed to every spoke to load/store data
                         from/to it
            :type data: pykickstart.base.BaseHandler
            :param storage: object storing storage-related information
                            (disks, partitioning, bootloader, etc.)
            :type storage: blivet.Blivet
            :param payload: object storing packaging-related information
            :type payload: pyanaconda.packaging.Payload
            :param instclass: distribution-specific information
            :type instclass: pyanaconda.installclass.BaseInstallClass
        '''
        EditTUISpoke.__init__(self, app, data, storage, payload, instclass)

        self.args = self.data.addons.org_scientificlinux_contexts
        self.name = self.args.name

        log.debug("%s: adding ksdata to kickstart object", self.name)
        self.args.ksdata = data

    @property
    def completed(self):
        ''' Context is never mandatory '''
        return True

    @property
    def status(self):
        ''' Show the hub our status '''
        return get_status(self.args)

    def apply(self):
        '''nothing needed here, values are set in the self.args tree'''
        pass

    def __menu_methods(self):
        '''
            we need a consistent way of evaluating which menu method was
            selected, so we generate the choices here.

            :returns: a list of menu options
            :rtype: list
        '''
        _methods = []
        _methods.append(_("Add"))
        remove = _("Remove")
        for context in self.args.enabled_contexts:
            _methods.append(remove + ': ' + context)
        return _methods

    def refresh(self, args=None):
        ''' format the window to display '''
        EditTUISpoke.refresh(self, args)

        _methods = self.__menu_methods()
        self._window += [TextWidget(_("Choose an action."))]
        text = [TextWidget(m) for m in _methods]

        def _prep(i, w):  #pylint: disable=invalid-name
            """ Mangle our text to make it look pretty on screen. """
            number = TextWidget("%2d)" % (i + 1))
            return ColumnWidget([(4, [number]), (None, [w])], 1)

        # gnarl and mangle all of our widgets so things look pretty on screen
        choices = [_prep(i, w) for i, w in enumerate(text)]

        displayed = ColumnWidget([(78, choices)], 1)
        self._window.append(displayed)

        return True

    def input(self, args, key):
        """ Handle the input; this decides the action we take. """
        try:
            num = int(key)
        except ValueError:
            return key

        if num == 1: # add
            newspoke = SpecifyNewContextSpoke(self.app, self.data, self.storage, self.payload, self.instclass)
            self.app.switch_screen_modal(newspoke)
            self.apply()
            self.close()
            return INPUT_PROCESSED
        else:        # a remove method
            _methods = self.__menu_methods()
            num = num - 1 # now valid for list index
            target = _methods[num].split(':', 1)[1].lstrip().rstrip()
            self.args.remove(target)

class SpecifyNewContextSpoke(EditTUISpoke):  #pylint: disable=too-many-public-methods,too-many-ancestors
    """ Specify a new context. """
    title = N_("Specify Context Options")
    category = SoftwareCategory

    edit_fields = [
        Entry(N_("Context URL"), "url", re.compile(".*$"), True),
        Entry(N_("Context Name"), "name", re.compile(".*$"), True),
        Entry(N_("Enable Testing"), "enable_testing", EditTUISpoke.CHECK, True)
        ]

    def __init__(self, app, data, storage, payload, instclass): #pylint: disable=too-many-arguments
        '''self.args should hold the object this Spoke is supposed to edit'''
        EditTUISpoke.__init__(self, app, data, storage, payload, instclass)

        self.args = Context()

    def refresh(self, args=None):
        """ Refresh window. """
        return EditTUISpoke.refresh(self, args)

    @property
    def indirect(self):
        return True

    def apply(self):
        """ Apply all of our changes. """
        self.data.addons.org_scientificlinux_contexts.add(self.args)
